/*
 * Solarlader.c
 *
 * Last update: 19-Apr-2014 
 * Author: Stephan.LaageWitt
 */ 

#define VERSION "1.05"

#define F_CPU 2000000

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/sleep.h>
#include "lcd-routines.c"

// Schaltspannungen und Strme
#define SOLAR_OFF_THRESHOLD 1420	// Solarpanel abschalten, wenn die Spannung ber diesen Wert steigt (in 10mV)
#define SOLAR_ON_THRESHOLD 1320		// Solarpanel einschalten, wenn die Spannung unter diesen Wert fllt (in 10mV)
#define BATTERY_LOW_THRESHOLD 1100	// schaltet alle Ausgnge aus, wenn die Spannung unter diesen Wert fllt (in 10mV)
#define POWER_OFF_COUNT 20			// Zeit fr Auto-Power-Off in Sekunden
#define POWER_OFF_V5 5				// Strom-Untergrenze fr 5V-Auto-Power-Off in 10mA
#define POWER_OFF_V12 50			// Strom-Untergrenze fr 12V-Auto-Power-Off in 10mA
#define MAX_I_5V 175				// Strom-Obergrenze fr 5V-Ausgang in 10mA
#define MAX_I_12V 400				// Strom-Obergrenze fr 12V-Ausgag in 10mA
#define SHUNT_IN 330				// Input-Shunt-Widerstand in mOhm
#define SHUNT_OUT_5V 220			// Output-Shunt-Widerstand in mOhm
#define SHUNT_OUT_12V 110			// Output-Shunt-Widerstand in mOhm
#define ADC_REFERENCE 1100			// Band-Gap Referenz-Spannung des Controllers
#define IRQ_PER_SEC 10 				// Anzahl der Interrupts pro Sekunde 

// Pin-Definitionen
#define pin_KEY_5VON 1				// Port D
#define pin_KEY_12VON 2				// Port D
#define pin_ADC_U5V_OUT 0			// Port C (ADC)
#define pin_ADC_12V_OUT 1			// Port C (ADC)
#define pin_ADC_U_SOL_IN 2			// Port C (ADC)
#define pin_ADC_U_BAT 3				// Port C (ADC)
#define pin_ADCSWITCH_5V_OUT 4		// Port C - U_5V_Out
#define pin_ADCSWITCH_5V_SHUNT 5	// Port C - U_5V_Shunt	 	
#define pin_SWITCH_SOL_IN 3			// Port D
#define pin_SWITCH_5V_OUT 4			// Port D
#define pin_SWITCH_12V_OUT 5		// Port D

// Globale Variablen 
uint32_t U_bat,	 					// Batterie-Spannung in 10mV
		 Cap_in,					// Input Ladekapazitt in 10mA-Stunden
		 Cap_5V_out,				// Output-Kapazitt in 10mA-Stunden 
		 Cap_12V_out;				// Output-Kapazitt in 10mA-Stunden
uint32_t I_in = 0,					// Solar-Strom in 10 mA
		 I_5V_out = 0,				// 5V-Ausgangs-Strom in 10 mA
		 I_12V_out = 0;				// 12V-Ausgangs-Strom in 10 mA
uint8_t	 hour = 0,					// Zeit, die die Ausgnge aktiv sind
		 minute = 0,
		 second = 0;
uint8_t	 V5_power_off_count = 0,	// Zhler fr 5V-Auto-Power-Off 
		 V12_power_off_count = 0;	// Zhler fr 12V-Auto-Power-Off
		 
/* Interrupt Service Handler ---------------------------------------------------------------------*/
/* Weckt die CPU 10 mal pro Sekunden auf.										                  */
ISR(TIMER1_COMPA_vect) {
	;								// CPU aufwecken
}	

/* lcd_dec_unit32 --------------------------------------------------------------------------------*/
/* Gibt einen 32-Bit-Wert als Dezimalzahl aus.													  */
/* positions -> Anzahl der Stellen (2 ... 7)													  */
/* dec_point -> Anzahl der Nachkommastellen	(0 ... positions)									  */
void lcd_dec_uint32(uint32_t value, uint8_t positions, const uint8_t dec_point)
{
	uint8_t count, blank_flag = 0;
	uint32_t denominator = 10;

	for (count = 2; count < positions; ++count)
		denominator *= 10;
	 	
	while (positions > 0) {
		count = value / denominator;
		value -= count * denominator;
		if (positions == dec_point + 1) blank_flag = 1;
		if (positions == dec_point)	lcd_data('.');
		if ((count == 0) && (blank_flag == 0)) { 
			lcd_data(' ');
		} else {
			lcd_data('0' + count);
			blank_flag = 1;
		};			
		--positions;
		denominator /= 10;
	};					
}

/* lcd_two_digit  --------------------------------------------------------------------------------*/
/* Gibt einen 8-Bit-Wert als zweisetllig Dezimalzahl aus										  */
void lcd_two_digit(uint8_t number) 
{
	uint8_t dez = number / 10;
	
	lcd_data('0' + dez);
	number -= 10 * dez;
	lcd_data('0' + number);
} 


/* GetADC ----------------------------------------------------------------------------------------*/
/* Lsst den ADC fr einen gewhlten Kanal laufen						                  	      */ 
uint16_t GetAdc(const uint8_t channel) {
	ADMUX  = (1<<REFS1) | (0<<ADLAR) | channel;		// Kanal setzen			
	ADCSRA |= (1<<ADSC);							// ADC starten 
	while ( ADCSRA & (1<<ADSC) );					// ... auf den ADC-Wert warten
	return(ADCW);									// ADC-Wert zurck geben
}
	
/* ShowData --------------------------------------------------------------------------------------*/
/* betreibt das Display																			  */
void ShowData(void) {

	// U_bat
	lcd_setcursor(0, 1);
	lcd_string("Bat ");
	lcd_dec_uint32(U_bat, 4, 2);
	lcd_data('V');

	// I_in
	lcd_setcursor(2, 2);
	lcd_string("In ");
	if (PORTD & (1<<pin_SWITCH_SOL_IN)) {
		lcd_string("-.-- ");
	} else {
		lcd_dec_uint32(I_in, 3, 2);
		lcd_data('A');
	};

	// Cap_in
	lcd_setcursor(1, 3);
	lcd_dec_uint32(Cap_in / 360, 6, 3);
	lcd_data('A');
	lcd_data('h');

	// Output Zeit
	lcd_setcursor(2, 4);
	lcd_two_digit(hour);
	lcd_data(':');
	lcd_two_digit(minute);
	lcd_data(':');
	lcd_two_digit(second);		

	// U_5V_out
	lcd_setcursor(11, 1);
	lcd_string("USB ");
	if (PORTD & (1<<pin_SWITCH_5V_OUT)) {
		lcd_string("-.-- ");
	} else {
		lcd_dec_uint32(I_5V_out, 3, 2);
		lcd_data('A');
	};
	lcd_setcursor(11, 2);
	lcd_dec_uint32(Cap_5V_out / 360, 6, 3);
	lcd_data('A'); 
	lcd_data('h');
	
	// 12 V Output
	lcd_setcursor(11, 3);
	lcd_string("12V ");
  	if (PORTD & (1<<pin_SWITCH_12V_OUT)) {
  		lcd_string("-.-- ");
  	} else {
  		lcd_dec_uint32(I_12V_out, 3, 2);
	  	lcd_data('A');
  	};
  	lcd_setcursor(11, 4);
  	lcd_dec_uint32(Cap_12V_out / 360, 6, 3);
  	lcd_data('A');
  	lcd_data('h');
}

/* Set Charge Mode ------------------------------------------------------------------------------*/
/* schaltet das Solarpanel ein oder aus															 */
void set_charge_mode(const uint8_t status) {
	if (status) {
		PORTD &= ~(1<<pin_SWITCH_SOL_IN);			// Solarzelle einschalten 
	} else {
		PORTD |= (1<<pin_SWITCH_SOL_IN);			// Solarzelle abschalten 	
	}
}

/* Set 5V Output Mode ------------------------------------------------------------------------------*/
/* Schaltet den 5V-Konverter ein oder aus															*/
void set_5V_output_mode(const uint8_t status) {
	if (status) {
		PORTD &= ~(1<<pin_SWITCH_5V_OUT);			// USB Output einschalten 
		Cap_5V_out = 0;								// Stromkapazitts-Zhler zurcksetzen
		hour = minute = second = 0;					// setze Zeit auf 0
		V5_power_off_count = POWER_OFF_COUNT; 
	} else {
		PORTD |= (1<<pin_SWITCH_5V_OUT);			// 5V-USB Output abschalten 	
	}
}

/* Set 12V Output Mode ------------------------------------------------------------------------------*/
/* schaltet den 12V-Ausgang ein oder aus															 */
void set_12V_output_mode(const uint8_t status) {
	if (status) {
		PORTD &= ~(1<<pin_SWITCH_12V_OUT);			// 12V-Ausgang einschalten 
		Cap_12V_out = 0;							// Stromkapazitts-Zhler zurcksetzen
		hour = minute = second = 0;					// setze Zeit auf 0
		V12_power_off_count = POWER_OFF_COUNT;
	} else {
		PORTD |= (1<<pin_SWITCH_12V_OUT);			// 12V-Ausgang abschalten	
	}
}

/* --------------------------------------------------------------------------------------------------*/
int main(void)
{
	uint8_t irq_count = 0;			// counts IRQ upwards
	uint32_t U_solar_shunt, 		// Spannung am Solar-Shunt in mV
			 U_12V_shunt;			// Spannung am 12V-Shunt in mV
	uint32_t adc_bat = 0, adc_solar_shunt = 0, adc_5V_out = 0, adc_5V_shunt = 0, adc_12V_shunt = 0;
	
	// Ports
	DDRC = 0b11110000;				// PC0 = ADC0 = adc_5V_shunt und adc_5V_out 
									// PC1 = ADC1 = adc_12V_shunt
									// PC2 = ADC2 = adc_solar
									// PC3 = ADC3 = adc_bat
									// PC4 = Output, U_5V_shunt_out aktiv 
									// PC5 = Output, U_5V_out aktiv
	DDRD = 0b11111000;				// PD1: input, Taste fr 5V ein  (low -> active)
									// PD2: input, Taste fr 12V ein (low -> active)
									// PD3: output, schaltet Solarpanel ein (low -> active)
									// PD4: output, schaltet 5V-Out ein (low -> active)
									// PD5: output, schaltet 12-Out ein (low -> active)
 	PORTC = 0b11110000;
	PORTD = 0b11111111;
	
	// ADC initialisieren 
	ADMUX  = (1<<REFS0) | (0<<ADLAR);								// 10 bit, right adjusted				
	ADCSRA = (1<<ADEN) | (1<<ADPS2) | (0<<ADPS1) | (1<<ADPS0);		// scaling factor: 32 -> 62.500 kHz 
	
	// LCD initialisieren
	lcd_init();
	lcd_string("Sol v");
	lcd_string(VERSION);											// Software Versionsnummer ausgeben

	// Sleep Mode setzen: Interrupts unterbrechen den Sleep-Mode 
	set_sleep_mode(SLEEP_MODE_IDLE);

	// Timer1 starten, Interrupt 10 mal pro Sekunde 
	OCR1A = 25000-1;											  
	TCCR1B |= (1<<WGM12)| (0<<CS02) | (1<<CS01) |(0<<CS00);  		// CPU-Frequenz / 8, CTC1 = Clear Timer on Compare
	TIMSK1 |= (1 << OCIE1A); 	
	sei();															// Interrup aktiv schalten

    while(1)
    {
		// ADC-Werte einlesen ----------------------------------------------------------------------
		ADCSRA |= (1<<ADEN);							// ADC einschalten 
		GetAdc(0);										// den ersten Wert nach dem Einschalten verwerfen			

		adc_bat += GetAdc(pin_ADC_U_BAT);				// Batteriespannung einlesen				
		adc_solar_shunt += GetAdc(pin_ADC_U_SOL_IN);	// Spannung am Solar-Shunt einlesen	
		adc_12V_shunt += GetAdc(pin_ADC_12V_OUT);		// Spannung am 12V-Shunt einlesen
	
		PORTC &= ~(1<<pin_ADCSWITCH_5V_OUT);			// Schalter auf 5V-Ausgangs-Spannung setzen
		GetAdc(0);										// warte einen Moment, Wert wird verworfen
		adc_5V_shunt += GetAdc(pin_ADC_U5V_OUT);		// 5V-Ausgangsspannung einlesen
		PORTC |= (1<<pin_ADCSWITCH_5V_OUT);				// Schalter zurcksetzen
	
		PORTC &= ~(1<<pin_ADCSWITCH_5V_SHUNT);			// Schalter auf 5-V-Shunt-Spannung setzen 
		GetAdc(0);										// warte einen Moment, Wert wird verworfen
		adc_5V_out += GetAdc(pin_ADC_U5V_OUT);			// 5V-Shunt-Spannung einlesen
		PORTC |= (1<<pin_ADCSWITCH_5V_SHUNT);			// Schalter zurcksetzen
	
		ADCSRA &= ~(1<<ADEN);							// ADC abschalten, um Strom zu sparen 

		// Tasten abfragen ----------------------------------------------------------------------
		if (!(PIND & (1<<pin_KEY_5VON))) {				// Taste fr 5V-USB gedrckt
			set_5V_output_mode(1);						// 5V-USB einschalten
		};
		if (!(PIND & (1<<pin_KEY_12VON))) {				// Taste fr 12V gedrckt
			set_12V_output_mode(1);						// 12V einschalten
		};
		
		// dieser Zweig wird einmal pro Sekunde durchlfaufen -------------------------------------
		if (++irq_count == IRQ_PER_SEC) {

			// Ladezeit heraufzhlen wenn einer der Ausgnge aktiv ist
			if (!(PORTD & (1<<pin_SWITCH_5V_OUT)) || !(PORTD & (1<<pin_SWITCH_12V_OUT))) {
				if (++second == 60) {
					second = 0;
					if (++minute == 60) {
						minute = 0;
						++ hour;
					};
				};
			}; 
			
			// Solar-Spannung und Strom berechnen
			U_solar_shunt = adc_solar_shunt * ADC_REFERENCE / (1024 * IRQ_PER_SEC);		// Spannung in mV
			I_in = U_solar_shunt * 100 / SHUNT_IN;										// Solarstrom berechnen
			Cap_in = Cap_in + I_in;														// Solar-Ladekapazitt berechnen
			
			// Batterie-Spannung berechnen
			U_bat = (adc_bat * 15 / IRQ_PER_SEC - U_solar_shunt) / 10;					// Spannung in 10mv
	
			// 5V-USB-Strom berechnen wenn 5V eingeschaltet ist und U_5V_shunt > U_5V_out
			if (!(PORTD & (1<<pin_SWITCH_5V_OUT))) {
				if (adc_5V_shunt > (adc_5V_out + 5)) {   
					I_5V_out = (adc_5V_shunt - adc_5V_out) * ADC_REFERENCE / (1024 * IRQ_PER_SEC);
					I_5V_out = I_5V_out * 100 / SHUNT_OUT_5V;
					Cap_5V_out = Cap_5V_out + I_5V_out;
				} else {
					I_5V_out = 0;
				};					
			};								

			// 12V Strom berechnen
			U_12V_shunt = adc_12V_shunt * ADC_REFERENCE / (1024 * IRQ_PER_SEC);			// Spannung in mV
			if (U_12V_shunt > U_solar_shunt) {
				U_12V_shunt -= U_solar_shunt;
				I_12V_out = U_12V_shunt * 100 / SHUNT_OUT_12V;
				Cap_12V_out = Cap_12V_out + I_12V_out;
			} else {
				I_12V_out = 0;
			};			
			
			// Alle Daten anzeigen	
			ShowData();								

			// setze Solarpanel-Schalter 
			if (U_bat > SOLAR_OFF_THRESHOLD) { 
				set_charge_mode(0); 		// schalte das Solarpanel ab 
			};					
			if (U_bat < SOLAR_ON_THRESHOLD) {
				set_charge_mode(1);			// schalte das Solarpanel ein
			};				
			
			// Auto-Power-Off fr 5V-USB
			if  (!(PORTD & (1<<pin_SWITCH_5V_OUT))) {
				if (I_5V_out < POWER_OFF_V5) {
					if (V5_power_off_count > 0) {
						--V5_power_off_count;
					} else {
						set_5V_output_mode(0);
					};
				} else {
					V5_power_off_count = POWER_OFF_COUNT;
				};										
			};
			
			// Auto-Power-Off fr 12V-Ausgang
			if (!(PORTD & (1<<pin_SWITCH_12V_OUT))) {
				if (I_12V_out < POWER_OFF_V12) { 
					if (V12_power_off_count > 0) {
						--V12_power_off_count;
					} else {
					set_12V_output_mode(0);
					};
				} else {
					V12_power_off_count = POWER_OFF_COUNT;		
				};					
			};
			
			// Strom-Obergrenze fr 5V-Ausgang testen
			if (I_5V_out > MAX_I_5V)
				set_5V_output_mode(0);

			// Strom-Obergrenze fr 12V-Ausgang testen
			if (I_12V_out > MAX_I_12V)
				set_12V_output_mode(0);
				
			// Unterspannung testen
			if (U_bat < BATTERY_LOW_THRESHOLD) {
				set_5V_output_mode(0);
				set_12V_output_mode(0);
			}
			
			// Variablen zurcksetzen 
			irq_count = 0;
			adc_solar_shunt = 0;
			adc_bat = 0;	
			adc_5V_out = 0;
			adc_5V_shunt = 0;		
			adc_12V_shunt = 0;	
		};  
		
		// Stromsparen		
		sleep_mode();
		
    }; // closing while (1)
}	
